"""Functions to call system programs and python modulse."""

import logging
import os
import subprocess

from incipyt import project

from incipyt._internal.templates import Formattable
from incipyt._internal.utils import EnvValue

logger = logging.getLogger(__name__)


def run(args, check=True, **kwargs):
    r"""Run a command after substitution using the environ.

    :param args: List of the command elements.
    :type args: :class:`list`
    :param \**kwargs: Other options forwarded to `subprocess.run`
    :return: Represents a process that has finished
    :rtype: :class:`subprocess.CompletedProcess`
    """
    formatted = [arg.format() if isinstance(arg, Formattable) else arg for arg in args]
    logger.info(" ".join(formatted))
    result = subprocess.run(formatted, capture_output=True, check=False, **kwargs)
    logger.info(result.stdout.decode())
    if check and result.returncode:
        logger.error(result.stderr.decode())
        raise subprocess.CalledProcessError(
            result.returncode, result.args, output=result.stdout, stderr=result.stderr
        )
    return result


def setenv_python_cmd(python_path):
    """Set PYTHON_CMD environment variable.

    :param python_path: List of the command elements.
    :type python_path: :class:`pathlib.Path`
    """
    assert python_path.is_absolute(), f"{python_path} is not absolute."
    project.environ["PYTHON_CMD"] = EnvValue(os.fspath(python_path), update=True)


def python_m(args, **kwargs):
    r"""Run a python module after substitution using the environ.

    :param args: List of the command elements, excluding `python -m`.
    :type args: :class:`list`
    :param \**kwargs: Other options forwarded to `subprocess.run`
    :return: Represents a process that has finished
    :rtype: :class:`subprocess.CompletedProcess`
    """
    # from incipyt._internal.templates import StringTemplate
    return run([project.environ["PYTHON_CMD"], "-m"] + args, **kwargs)


def build(args, **kwargs):
    r"""Run a python build after substitution using the environ.

    :param args: List of the command elements, excluding `python -m build`.
    :type args: :class:`list`
    :param \**kwargs: Other options forwarded to `subprocess.run`
    :return: Represents a process that has finished
    :rtype: :class:`subprocess.CompletedProcess`
    """
    return python_m(["build"] + args, **kwargs)


def pip(args, **kwargs):
    r"""Run a pip command after substitution using the environ.

    :param args: List of the command elements, excluding `python -m pip`.
    :type args: :class:`list`
    :param \**kwargs: Other options forwarded to `subprocess.run`
    :return: Represents a process that has finished
    :rtype: :class:`subprocess.CompletedProcess`
    """
    return python_m(["pip", "--verbose"] + args, **kwargs)


def pip_install(args, **kwargs):
    r"""Run a pip install command after substitution using the environ.

    :param args: List of the command elements, excluding `python -m pip`.
    :type args: :class:`list`
    :param \**kwargs: Other options forwarded to `subprocess.run`
    :return: Represents a process that has finished
    :rtype: :class:`subprocess.CompletedProcess`
    """
    return pip(["install", "--upgrade", "--upgrade-strategy", "eager"] + args, **kwargs)


def venv(args, **kwargs):
    r"""Run a venv command after substitution using the environ.

    :param args: List of the command elements, excluding `python -m venv`.
    :type args: :class:`list`
    :param \**kwargs: Other options forwarded to `subprocess.run`
    :return: Represents a process that has finished
    :rtype: :class:`subprocess.CompletedProcess`
    """
    return python_m(["venv"] + args, **kwargs)



import click
from pytest import fixture, mark, raises

from incipyt._internal.templates import (
    ChoiceTemplate,
    StringTemplate,
    TemplateDict,
)
from incipyt import project
from tests.utils import mock_stdin


class TestStringTemplate:
    @fixture
    def reset_environ(self):
        project.environ.clear()

    @fixture
    def simple_st(self):
        return StringTemplate("{ONE}")

    @fixture
    def kwarg_st(self):
        return StringTemplate("{ONE}", ONE="1-kwarg")

    @fixture
    def sanitizer_st(self):
        return StringTemplate("{ONE}", sanitizer=lambda k, v: f"{v}-sanitizer")

    @fixture
    def multiple_st(self):
        return StringTemplate("{ONE}-{TWO}-{THREE}")

    def test_env_key_push(self, kwarg_st, reset_environ, monkeypatch):
        mock_stdin(monkeypatch, "")
        kwarg_st.format()
        assert project.environ["ONE"] == "1-kwarg"

    def test_env_key_push_prompt(self, simple_st, reset_environ, monkeypatch):
        mock_stdin(monkeypatch, "1")
        simple_st.format()
        assert project.environ["ONE"] == "1"

    @mark.parametrize(
        "st, variables, stdin, res",
        (
            ("simple_st", {"ONE": "1"}, "", "1"),
            ("kwarg_st", {}, "", "1-kwarg"),
            ("sanitizer_st", {"ONE": "1"}, "", "1-sanitizer"),
            ("multiple_st", {"ONE": "1", "TWO": "2", "THREE": "3"}, "\n\n", "1-2-3"),
        ),
    )
    def test_format(
        self, st, variables, stdin, res, reset_environ, request, monkeypatch
    ):
        mock_stdin(monkeypatch, stdin)
        st = request.getfixturevalue(st)
        project.environ |= variables
        assert st.format() == res

    @mark.parametrize(
        "st, stdin",
        (
            ("simple_st", ""),
            ("multiple_st", "\n\n"),
        ),
    )
    def test_format_null(self, st, stdin, reset_environ, request, monkeypatch):
        mock_stdin(monkeypatch, stdin)
        st = request.getfixturevalue(st)
        project.environ |= {"ONE": "", "TWO": "2", "THREE": "3"}
        assert st.format() is None


class TestChoiceTemplate:
    @fixture
    def simple_mst(self):
        return ChoiceTemplate("a", "b")

    @fixture
    def formattable_mst(self):
        return ChoiceTemplate(StringTemplate("a"), StringTemplate("b"))

    @fixture
    def reset_environ(self):
        project.environ.clear()

    def test_mst_tail(self, simple_mst):
        mst = ChoiceTemplate("x", simple_mst)
        assert mst._values == {
            StringTemplate("x"),
            StringTemplate("a"),
            StringTemplate("b"),
        }

    @mark.parametrize("mst", ("simple_mst", "formattable_mst"))
    def test_call(self, mst, reset_environ, monkeypatch, request):
        mock_stdin(monkeypatch, "a")
        mst = request.getfixturevalue(mst)
        assert mst.format() == "a"

    @mark.parametrize("mst", ("simple_mst", "formattable_mst"))
    def test_call_invalid(self, mst, reset_environ, monkeypatch, request):
        mock_stdin(monkeypatch, "x")
        mst = request.getfixturevalue(mst)
        with raises(click.exceptions.Abort):
            mst.format()


class TestTemplateCollection:
    @fixture
    def empty_td(self):
        return TemplateDict({})

    @fixture
    def simple_td(self):
        return TemplateDict({"1": StringTemplate("a")})

    @fixture
    def nested_td(self):
        return TemplateDict({"1": {"2": {"3": StringTemplate("a")}}})

    @fixture
    def choice_td(self):
        return TemplateDict({"1": ChoiceTemplate("a", "b")})

    @fixture
    def sequence_td(self):
        return TemplateDict({"1": [StringTemplate("a"), StringTemplate("b")]})

    @mark.parametrize(
        "td, res",
        (
            ("empty_td", TemplateDict({"1": StringTemplate("x")})),
            (
                "simple_td",
                TemplateDict({"1": ChoiceTemplate("x", "a")}),
            ),
            (
                "choice_td",
                TemplateDict({"1": ChoiceTemplate.from_items("x", "a", "b")}),
            ),
        ),
    )
    def test_setitem(self, td, res, request):
        td = request.getfixturevalue(td)
        td["1"] = "x"
        assert td == res

    @mark.parametrize(
        "td, res",
        (
            ("empty_td", TemplateDict({"1": StringTemplate("x")})),
            (
                "simple_td",
                TemplateDict({"1": ChoiceTemplate("x", "a")}),
            ),
            (
                "choice_td",
                TemplateDict({"1": ChoiceTemplate.from_items("x", "a", "b")}),
            ),
        ),
    )
    def test_setitem_callable(self, td, res, request):
        td = request.getfixturevalue(td)
        td["1"] = StringTemplate("x")
        assert td == res

    @mark.parametrize(
        "td, res",
        (
            ("empty_td", TemplateDict({"1": {"2": {"3": StringTemplate("x")}}})),
            (
                "nested_td",
                TemplateDict({"1": {"2": {"3": ChoiceTemplate("x", "a")}}}),
            ),
        ),
    )
    def test_chained_setitem(self, td, res, request):
        td = request.getfixturevalue(td)
        td["1", "2", "3"] = "x"
        assert td == res

    @mark.parametrize(
        "td, res",
        (
            (
                "empty_td",
                TemplateDict({"1": [StringTemplate("a"), StringTemplate("x")]}),
            ),
            (
                "sequence_td",
                TemplateDict(
                    {
                        "1": [
                            StringTemplate("a"),
                            StringTemplate("b"),
                            StringTemplate("x"),
                        ]
                    }
                ),
            ),
        ),
    )
    def test_sequence_setitem(self, td, res, request):
        td = request.getfixturevalue(td)
        td["1"] = ["a", "x"]
        assert td == res

    @mark.parametrize(
        "td, res",
        (
            (
                "empty_td",
                TemplateDict(
                    {
                        "1": [
                            StringTemplate("a"),
                            [StringTemplate("x"), StringTemplate("y")],
                            {"2": StringTemplate("z")},
                        ]
                    }
                ),
            ),
            (
                "sequence_td",
                TemplateDict(
                    {
                        "1": [
                            StringTemplate("a"),
                            StringTemplate("b"),
                            [StringTemplate("x"), StringTemplate("y")],
                            {"2": StringTemplate("z")},
                        ]
                    }
                ),
            ),
        ),
    )
    def test_nested_sequence_setitem(self, td, res, request):
        td = request.getfixturevalue(td)
        td["1"] = ["a", ["x", "y"], {"2": "z"}]
        assert td == res

    @mark.parametrize(
        "td, res",
        (
            ("empty_td", TemplateDict({"1": {"2": {"3": StringTemplate("x")}}})),
            (
                "nested_td",
                TemplateDict({"1": {"2": {"3": ChoiceTemplate("x", "a")}}}),
            ),
        ),
    )
    def test_ior(self, td, res, request):
        td = request.getfixturevalue(td)
        td |= {"1": {"2": {"3": "x"}}}
        assert td == res


class TestTemplateVisitor:
    @fixture
    def reset_environ(self):
        project.environ.clear()

    @fixture
    def empty_td(self):
        return {}

    @fixture
    def simple_td(self):
        return {"1": StringTemplate("{ONE}")}

    @fixture
    def nested_td(self):
        return {"1": {"2": {"3": StringTemplate("{ONE}")}}}

    @fixture
    def choice_td(self):
        return {"1": ChoiceTemplate("{ONE}", "b")}

    @fixture
    def sequence_td(self):
        return {"1": [StringTemplate("{ONE}"), {2: StringTemplate("b")}]}

    @fixture
    def single_td(self):
        return {"1": [StringTemplate("{ONE}")]}

    @mark.parametrize(
        "td, res, input_values",
        (
            ("empty_td", {}, []),
            ("simple_td", {"1": "a"}, ["a"]),
            ("simple_td", {}, [""]),
            ("nested_td", {"1": {"2": {"3": "a"}}}, ["a"]),
            ("nested_td", {}, [""]),
            ("sequence_td", {"1": ["a", {2: "b"}]}, ["a"]),
            ("sequence_td", {"1": [{2: "b"}]}, [""]),
            ("single_td", {"1": ["a"]}, ["a"]),
            ("single_td", {}, [""]),
            ("choice_td", {"1": "a"}, ["a", "a"]),
        ),
    )
    def test_call(self, td, res, reset_environ, input_values, monkeypatch, request):
        mock_stdin(monkeypatch, input_values)
        td = request.getfixturevalue(td)
        project._Structure._visit(td)
        assert td == res



from pytest import fixture, mark, raises

from incipyt import project
from incipyt._internal.templates import FormatterEnviron

from tests.utils import mock_stdin


class _Context:
    @fixture
    def reset_environ(self):
        project.environ.clear()
        project.environ["VARIABLE_NAME"] = project.EnvValue("value", confirmed=True)
        project.environ["EMPTY_VARIABLE"] = project.EnvValue("", confirmed=True)

    @fixture
    def simple_ctx(self, reset_environ):
        return FormatterEnviron()

    @fixture
    def no_error_ctx(self, reset_environ):
        return FormatterEnviron(value_error=False)

    @fixture
    def populated_ctx(self, simple_ctx):
        simple_ctx.format("{VARIABLE_NAME}")
        return simple_ctx


class TestFormatterEnviron(_Context):
    def test_contains(self, simple_ctx):
        assert "VARIABLE_NAME" in simple_ctx

    @mark.parametrize("ctx", ("simple_ctx", "no_error_ctx"))
    def test_contains_undefined(self, ctx, monkeypatch, request):
        mock_stdin(monkeypatch, "value")
        ctx = request.getfixturevalue(ctx)
        assert "OTHER_NAME" in ctx
        assert ctx["OTHER_NAME"] == "value"

    def test_getitem_empty_simple(self, simple_ctx):
        with raises(ValueError):
            simple_ctx["EMPTY_VARIABLE"]

    def test_getitem_empty_no_error(self, no_error_ctx):
        assert no_error_ctx["EMPTY_VARIABLE"] == ""

    @mark.parametrize("ctx", ("simple_ctx", "no_error_ctx"))
    def test_getitem_sanitized(self, ctx, request):
        ctx = request.getfixturevalue(ctx)
        ctx._sanitizer = lambda k, v: v.upper()
        assert ctx["VARIABLE_NAME"] == "VALUE"

    @mark.parametrize(
        "ctx, res",
        (
            ("simple_ctx", set()),
            ("no_error_ctx", set()),
            ("populated_ctx", {"VARIABLE_NAME"}),
        ),
    )
    def test_iteration(self, ctx, res, request):
        ctx = request.getfixturevalue(ctx)
        assert set(ctx) == res
        assert len(ctx) == len(res)

    @mark.parametrize(
        "ctx, keys, values",
        (
            ("simple_ctx", [], []),
            ("no_error_ctx", [], []),
            ("populated_ctx", ["VARIABLE_NAME"], ["value"]),
        ),
    )
    def test_key_values(self, ctx, keys, values, request):
        ctx = request.getfixturevalue(ctx)
        assert list(ctx.keys()) == keys
        assert list(ctx.values()) == values
        assert list(ctx.items()) == list(zip(keys, values))


class TestRenderString(_Context):
    @mark.parametrize("ctx", ("simple_ctx", "no_error_ctx"))
    def test_interp(self, ctx, request):
        ctx = request.getfixturevalue(ctx)
        assert ctx.format("{VARIABLE_NAME}") == "value"

    @mark.parametrize("ctx", ("simple_ctx", "no_error_ctx"))
    def test_interp_kwarg(self, ctx, monkeypatch, request):
        mock_stdin(monkeypatch, "")
        ctx = request.getfixturevalue(ctx)
        assert ctx.format("{OTHER_NAME}", OTHER_NAME="value") == "value"

    @mark.parametrize("ctx", ("simple_ctx", "no_error_ctx"))
    def test_interp_undefined(self, ctx, monkeypatch, request):
        mock_stdin(monkeypatch, "value")
        ctx = request.getfixturevalue(ctx)
        assert ctx.format("{OTHER_NAME}") == "value"

    @mark.parametrize(
        "ctx, res",
        (
            ("simple_ctx", None),
            ("no_error_ctx", "value"),
        ),
    )
    def test_interp_empty(self, ctx, res, request):
        ctx = request.getfixturevalue(ctx)
        assert ctx.format("{VARIABLE_NAME}{EMPTY_VARIABLE}") == res



import logging
import os
import shutil
import sys

from incipyt import commands, project, signals, tools
from incipyt._internal.dumpers import TextFile


logger = logging.getLogger(__name__)


class Git(tools.Tool):
    """Scripts to add Git to :class:`incipyt.project._Structure`."""

    def __init__(self):
        if not shutil.which("git"):
            logger.error("%r is missing from the PATH. Abort.", "git")
            sys.exit(1)

        signals.vcs_ignore.connect(self._slot)

    def add_to_structure(self):
        """Add git configuration to `project.structure`.

        Register git related project URLs.

        URLs:
            - Repository: {REPOSITORY}
            - Issue: {REPOSITORY}/issues
            - Documentation: {REPOSITORY}/wiki
        """
        signals.project_url.emit(
            url_kind="Documentation", url_value="{REPOSITORY}/wiki"
        )
        signals.project_url.emit(url_kind="Issue", url_value="{REPOSITORY}/issues")
        signals.project_url.emit(url_kind="Repository", url_value="{REPOSITORY}")

    def _slot(self, pattern, **kwargs):
        project.structure.get_config_list(TextFile(".gitignore")).append(pattern)

    def pre(self, workon):
        """Run `git init`.

        Also push {AUTHOR_NAME} and {AUTHOR_EMAIL} using `git config user.*`.

        :param workon: Work-on folder.
        :type workon: :class:`pathlib.Path`
        """
        commands.run(["git", "init", os.fspath(workon)])

        project.environ["AUTHOR_EMAIL"] = project.EnvValue(
            commands.run(["git", "-C", os.fspath(workon), "config", "user.email"])
            .stdout.decode()
            .strip(),
            update=True,
        )
        project.environ["AUTHOR_NAME"] = project.EnvValue(
            commands.run(["git", "-C", os.fspath(workon), "config", "user.name"])
            .stdout.decode()
            .strip(),
            update=True,
        )

    def post(self, workon):
        """Run `git add --all`.

        :param workon: Work-on folder.
        :type workon: :class:`pathlib.Path`
        """
        commands.run(["git", "-C", os.fspath(workon), "add", "--all"])



"""TO-DO."""

import collections
import logging
import os
import sys
from collections import abc

import click

from incipyt._internal import templates

from incipyt._internal.utils import EnvValue, is_nonstring_sequence

logger = logging.getLogger(__name__)


class _Environ(collections.UserDict):
    """Manage environ variables using for substitutions in patterns.

    A :class:`Environ` object `environ` is a dictionnary initialized with
    system environ variables `os.environ`, a variable can be add like that:

    .. code-block::

        environ["VARIABLE_NAME"] = "something"

    However such a variable will not be considered in `environ` before being
    `confirmed`. Such confirmation will happen the first it is asked for:

    .. code-block::

        var_first = environ["VARIABLE_NAME"]
        # Prompt a message for confirmation of the value for `VARIABLE_NAME`
        var_second = environ["VARIABLE_NAME"]
        # As `VARIABLE_NAME` as been confirmed, don't ask anything

    Moreover, even if a specific variable hasn't been set, it can be asked for
    it anyway and it will be confirmed immediately. A variable can also be set
    and confirmed at the same time:

    .. code-block::

        environ["VARIABLE_NAME"] = EnvValue("something", confirmed=True)

    Note that if a variable has been set, it cannot be set again, except if
    explicitly specified:

    .. code-block::

        environ["VARIABLE_NAME"] = EnvValue("new_value", update=True)

    All iterative methods and in operator consider only `confirmed` variables.
    """

    def clear(self):
        self._confirmed = set()
        self.data = os.environ.copy()

        if "PYTHON_CMD" not in self.data:
            self.data["PYTHON_CMD"] = sys.executable
        self._confirmed.add("PYTHON_CMD")

    def __init__(self):
        self.clear()

    def __getitem__(self, key):
        if key not in self._confirmed:
            logger.debug("Missing environ variable %s, request it.", key)
            self.data[key] = self._prompt(key)
            self._confirmed.add(key)

        return self.data[key]

    def __setitem__(self, key, env_value):
        if not isinstance(env_value, EnvValue):
            env_value = EnvValue(env_value)

        if key in self.data and not env_value.update:
            raise ValueError(f"Environ variable {key} already exists, use update.")

        logger.debug("Set environ variable %s=%s.", key, env_value.value)
        self.data[key] = env_value.value
        if env_value.confirmed and key not in self._confirmed:
            self._confirmed.add(key)

    def __iter__(self):
        return iter(self._confirmed)

    def __contains__(self, key):
        return key in self._confirmed

    def keys(self):  # noqa: D102
        return self._confirmed

    def values(self):  # noqa: D102
        return [self.data[key] for key in self._confirmed]

    def items(self):  # noqa: D102
        return [(key, self.data[key]) for key in self._confirmed]

    def getitems_sanitized(self, keys, sanitizer=None):
        """Get multiple items at once and sanitize them.

        See also :func:`incipyt.system.Environment.__getitem__`, which will be
        used to pull each key from the environment.

        :param keys: Required environment keys. If a key is `None`, it will be ignored.
        :type keys: :class:`collections.abc.Sequence`
        :param sanitizer: Will be called on key-value pairs to sanitize values.
        :type sanitizer: :class:`function`
        :return: Sanitized environment key-value pairs.
        :rtype: :class:`dict`
        """
        return {
            key: sanitizer(key, self[key]) if sanitizer else self[key]
            for key in keys
            if key is not None
        }

    def _prompt(self, key):
        return click.prompt(
            key.replace("_", " ").lower().capitalize(),
            default=self.data[key] if key in self.data else "",
            type=str,
        )

    def __ior__(self, other):
        self.update(other)
        return self


environ = _Environ()


class _Structure:
    """Represents all configuration and template files to commit for the new project.

    An instance internally stores mappables between path objects and template
    files or dictionaries modeling configuration files.

    Functions :meth:`get_configuration` and :meth:`register_template` add
    respectively configuration dictionary and template to the instance.

    When the project structure is finally ready, functions :meth:`mkdir` :meth:`commit`
    can be used to write folder and files in a new folder after substituting
    variables in path and files using an :class:`incipyt.project.Environ`.
    """

    def clear(self):
        self._configurations = {}

    def __init__(self):
        self.clear()

    def get_config_dict(self, config_root):
        """Get a configuration dictionary associated to the relative path `config_root`.

        :param config_root: Relative path of the configuration file.
        :type config_root: :class:`pathlib.Path`
        :return: A reference to the configuration dictionary
        :rtype: :class:`incipyt._internal.templates.TempateDict`
        """
        if config_root not in self._configurations:
            logger.debug(
                "Register configuration %s in project structure.", str(config_root)
            )
            self._configurations[config_root] = {}

        if not isinstance(self._configurations[config_root], abc.MutableMapping):
            raise TypeError(f"{config_root} is not a dict.")
        return templates.TemplateDict(self._configurations[config_root])

    def get_config_list(self, config_root):
        """Get a configuration list associated to the relative path `config_root`.

        :param config_root: Relative path of the configuration file.
        :type config_root: :class:`pathlib.Path`
        :return: A reference to the configuration list
        :rtype: :class:`incipyt._internal.templates.TempateList`
        """
        if config_root not in self._configurations:
            logger.debug(
                "Register configuration %s in project structure.", str(config_root)
            )
            self._configurations[config_root] = []

        if not is_nonstring_sequence(self._configurations[config_root]):
            raise TypeError(f"{config_root} is not a list.")
        return templates.TemplateList(self._configurations[config_root])

    def commit(self):
        """Commit current project structure on disk.

        :raises RuntimeError: If one of the configuration file already exists.
        """
        for config_root, config in self._configurations.items():
            logger.info("Process environ variables for %s.", str(config_root))
            _Structure._visit(config)

        for config_root, config in self._configurations.items():
            logger.info("Write configuration file %s.", str(config_root))
            config_root.dump_in(config)

    def mkdir(self, workon):
        """Make all directories of the project structure in workon.

        :param workon: Work-on path.
        :type workon: :class:`pathlib.Path`
        """
        for config_root in self._configurations:
            logger.debug("Commit %s path.", str(config_root))
            config_root.commit(workon)

        for config_root in self._configurations:
            logger.info("Mkdir folders for %s.", str(config_root))
            config_root.mkdir()

    @staticmethod
    def _visit(template):
        """Visit the `template` nested-dictionary structure.

        All :class:`incipyt._internal.templates.Formattable` values of the template dictionary will be
        evaluated and replaced by their results. All nested structures will be recursively
        visited and processed too.

        :param template: The template dictionary or list to visit.
        :type template: :class:`collections.abc.MutableMapping` of :class:`collections.abc.MutableSequence`
        """
        if is_nonstring_sequence(template):
            for index, value in enumerate(template):
                if isinstance(value, templates.Formattable):
                    template[index] = value.format()
                else:
                    _Structure._visit(value)
                if not template[index]:
                    template[index] = None

            while None in template:
                template.remove(None)

        elif isinstance(template, abc.MutableMapping):  # noqa: SIM106
            for key, value in template.items():
                logger.debug("Visit %s to process environ variables.", key)

                if isinstance(value, templates.Formattable):
                    template[key] = value.format()
                else:
                    _Structure._visit(value)
                if not template[key]:
                    template[key] = None

            for key in [key for key, value in template.items() if value is None]:
                del template[key]

        else:
            raise AssertionError(f"{type(template)} do not support visitation.")


structure = _Structure()



import os
import textwrap
import sys

from incipyt import commands, project, signals, tools
from incipyt._internal import sanitizers, templates
from incipyt._internal.dumpers import CfgIni, TextFile, Toml

# as of may 2022, the latest stable release of most bsd/linux
# distributions ship a python whoose verion is at least 3.9
LINUX_MIN_PYTHON_VERSION = (3, 9)


class Setuptools(tools.Tool):
    """Scripts to add Setuptools to :class:`incipyt.project._Structure`."""

    def __init__(self, check=False):
        self.check_build = check
        signals.build_dependency.connect(self._slot_dependency)
        signals.classifier.connect(self._slot_classifier)
        signals.project_url.connect(self._slot_url)

    def add_to_structure(self):
        """Add setuptools configuration to `project.structure`.

        :file:`pyptoject.toml`

        .. code-block::

            [build-system]
            build-backend = "setuptools.build_meta"
            requires = [ "setuptools", "wheel",]

        If this configuration cannot be populate like that, an error is raised.

        :file:`setup.cfg`

        .. code-block::

            [metadata]
            author_email = {AUTHOR_NAME} <{AUTHOR_EMAIL}>
            description = {SUMMARY_DESCRIPTION}
            long_description = file: README.md
            long_description_content_type = text/markdown
            maintainer_email = {AUTHOR_NAME} <{AUTHOR_EMAIL}>
            name = {PROJECT_NAME}
            version = {PACKAGE_VERSION}
            classifiers =
                Programming Language :: Python :: 3 :: Only

            [options]
            packages = {PROJECT_NAME}
            python_requires = >={AUDIENCE_PYTHON_VERSION}

            [options.package_data]
            * = {PACKAGE_DATA}/*

            [options.extras_require]
            dev =
                build
                pip

        Here key-value association is appended, if a key already exists the
        value is appended to the current one(s), the user will be asked to
        choose when commiting.

        :raises RuntimeError: If a build-system is already setup im pyproject.toml.
        """
        pyproject = project.structure.get_config_dict(Toml("pyproject.toml"))
        setup = project.structure.get_config_dict(CfgIni("setup.cfg"))

        if "build-system" in pyproject:
            raise RuntimeError("Build system already registered.")

        pyproject["build-system"] = {
            "build-backend": "setuptools.build_meta",
            "requires": ["setuptools", "wheel"],
        }

        setup["metadata"] = {
            "author_email": "{AUTHOR_NAME} <{AUTHOR_EMAIL}>",
            "description": "{SUMMARY_DESCRIPTION}",
            "long_description": "file: README.md",
            "long_description_content_type": "text/markdown",
            "maintainer_email": "{AUTHOR_NAME} <{AUTHOR_EMAIL}>",
            "name": templates.StringTemplate(
                "{PROJECT_NAME}",
                sanitizer=sanitizers.project,
            ),
            "version": templates.StringTemplate(
                "{PACKAGE_VERSION}",
                sanitizer=sanitizers.version,
                PACKAGE_VERSION="0.0.0",
            ),
        }

        setup["options"] = {
            "packages": templates.StringTemplate(
                "{PROJECT_NAME}",
                sanitizer=sanitizers.package,
            ),
            "python_requires": templates.StringTemplate(
                ">={AUDIENCE_PYTHON_VERSION}",
                sanitizer=sanitizers.version,
                AUDIENCE_PYTHON_VERSION="{0[0]}.{0[1]}".format(  # noqa: FS002
                    min(sys.version_info, LINUX_MIN_PYTHON_VERSION)
                ),
            ),
        }

        setup["options.package_data", "*"] = templates.StringTemplate(
            "{PACKAGE_DATA}/*", confirmed=True, PACKAGE_DATA="data"
        )

        project.structure.get_config_list(
            TextFile("{PROJECT_NAME}/__init__.py", sanitizer=sanitizers.package)
        ).append("")

        project.structure.get_config_list(TextFile("LICENSE", sep="\n\n")).append(
            "Copyright (c) {AUTHOR_NAME} <{AUTHOR_EMAIL}>"
        )

        project.structure.get_config_list(TextFile("README.md", sep="\n\n")).append(
            templates.StringTemplate(
                textwrap.dedent(
                    """\
                # {PROJECT_NAME}

                {SUMMARY_DESCRIPTION}

                ## Usage

                ## Contribute

                Copyright (c) {AUTHOR_NAME} <{AUTHOR_EMAIL}>"""
                ),
                value_error=False,
            )
        )

        project.structure.get_config_list(TextFile("setup.py")).append(
            textwrap.dedent(
                """\
            import setuptools

            setuptools.setup()"""
            )
        )

        signals.build_dependency.emit(dep_name="build")
        signals.build_dependency.emit(dep_name="pip")

        signals.classifier.emit(
            classifier="Programming Language :: Python :: 3 :: Only"
        )

        signals.vcs_ignore.emit(pattern="dist")
        signals.vcs_ignore.emit(pattern="*.egg-info")

    def _slot_classifier(self, classifier, **kwargs):
        setup = project.structure.get_config_dict(CfgIni("setup.cfg"))
        if ("metadata", "classifiers") not in setup:
            setup["metadata", "classifiers"] = []
        setup["metadata", "classifiers"].append(classifier)

    def _slot_dependency(self, dep_name, **kwargs):
        setup = project.structure.get_config_dict(CfgIni("setup.cfg"))
        if ("options.extras_require", "dev") not in setup:
            setup["options.extras_require", "dev"] = []
        setup["options.extras_require", "dev"].append(dep_name)

    def _slot_url(self, url_kind, url_value, **kwargs):
        project.structure.get_config_dict(CfgIni("setup.cfg"))[
            "metadata", "project_urls"
        ] = {url_kind: url_value}

    def post(self, workon):
        """Editable install and build for test.

        :param workon: Work-on folder.
        :type workon: :class:`pathlib.Path`
        """
        commands.pip_install(["--editable", f"{workon}[dev]"])
        if self.check_build:
            commands.pip_install(["build"])
            commands.build([os.fspath(workon)])



"""This module contains classes related to template definition and rendering.

Incipyt mainly uses two kinds of templates: bare template strings that behave
like builtin Python string formatting. It also provides special wrappers to
allow easier templating of collections data structures, such as dict-like
template objects.
"""

import contextlib
import logging
from abc import ABCMeta, abstractmethod
from collections import abc
from string import Formatter

import click

from incipyt import project
from incipyt._internal import utils

logger = logging.getLogger(__name__)


class Formattable(metaclass=ABCMeta):
    @abstractmethod
    def format(self):  # noqa: A003
        raise NotImplementedError


class StringTemplate(Formattable):
    """This class acts like a wrapper around a format string.

    When an instance is called, it renders the underlying format string using
    environ values and overrides.
    """

    def __init__(
        self, format_string, confirmed=False, sanitizer=None, value_error=True, **kwargs
    ):
        r"""This class acts like a wrapper around a format string.

        When an instance is called, it renders the underlying format string
        using environ values and overrides.

        :param format_string: A format string whose keyword argument will be substituted.
        :type format_string: :class:`str`
        :param confirmed: Confirmed status for new variables from keyword args.
        :type confirmed: :class:`bool`, optionnal
        :param sanitizer: An optionnal callable to sanitize the values given (key, value) pairs.
        :type sanitizer: :class:`function` or `None`, optionnal
        :param value_error: Empty values generate errors.
        :type value_error: :class:`bool`
        :param \**kwargs: Variables overrides that will be used for rendering and pushed to the environ.
            Automatically wrapped in :class:`incipyt._internal.utils.EnvValue` if needed.
        :type \**kwargs: :class:`str`, optionnal
        """
        self._confirmed = confirmed
        self._sanitizer = sanitizer
        self._value_error = value_error
        self._format_string = format_string
        self._kwargs = kwargs

    def __eq__(self, other):
        return utils.attrs_eq(
            self,
            other,
            "_format_string",
            "_confirmed",
            "_sanitizer",
            "_value_error",
            "_kwargs",
        )

    def __hash__(self):
        return utils.attrs_hash(
            self,
            "_format_string",
            "_confirmed",
            "_sanitizer",
            "_value_error",
            **self._kwargs,
        )

    def format(self):  # noqa: A003
        """Format the underlying format string using variables from the environ.

        :return: The formatted string.
        :rtype: :class:`str`
        """
        return FormatterEnviron(
            sanitizer=self._sanitizer, value_error=self._value_error
        ).format(
            self._format_string,
            **{
                key: (
                    value
                    if isinstance(value, utils.EnvValue)
                    else utils.EnvValue(value, confirmed=self._confirmed)
                )
                for key, value in self._kwargs.items()
            },
        )

    def __repr__(self):
        return utils.make_repr(
            self,
            format_string=self._format_string,
            confirmed=self._confirmed,
            sanitizer=self._sanitizer,
            value_error=self._value_error,
            kwargs=self._kwargs,
        )

    @classmethod
    def wrap(cls, value):
        return value if isinstance(value, Formattable) else cls(value)


class ChoiceTemplate(Formattable):
    """Class to hold multiple values for a single key.

    When an instance is called, the user will be asked to pick a value using
    the command line interface.
    """

    def __init__(self, head, tail):
        """Class to hold multiple string for a single key.

        When an instance is called, the user will be asked to pick a string
        using the command line interface.

        :param head: Entry to put a the head of the stack.
        :type tail: :class:str
        :param tail: Tail of the stack.
        :type tail: :class:`incipyt._intternal.templates.ChoiceTemplate` or any bare value
        """
        self._values = (
            {StringTemplate.wrap(head)} | tail._values
            if isinstance(tail, ChoiceTemplate)
            else {
                StringTemplate.wrap(head),
                StringTemplate.wrap(tail),
            }
        )

    def format(self):  # noqa: A003
        """Ask the user to pick a value using the command line interface.

        If it is :class:`incipyt._internal.templates.Formattable`, it will be formatted.

        :return: The user-choosen value.
        """
        return click.prompt(
            "Conflicting configuration, choose between",
            type=click.Choice(
                [
                    value.format() if isinstance(value, Formattable) else value
                    for value in self._values
                ]
            ),
        )

    def __eq__(self, other):
        return utils.attrs_eq(self, other, "_values")

    def __hash__(self):
        return hash(tuple(self._values))

    def __repr__(self):
        return f"{type(self).__name__}({self._values})"

    @classmethod
    def from_items(cls, *args):
        r"""Build a class instance from any number of bare values.

        :param \*args: Entries to wrap.
        :return: New class instance
        :rtype: :class:`incipyt._intternal.templates.ChoiceTemplate`
        """
        instance = cls.__new__(cls)
        instance._values = {StringTemplate.wrap(arg) for arg in args}
        return instance


class TemplateDict(abc.MutableMapping):
    """Proxy class around a provided mapping.

    This is itended to ease configuration templating.

    :Class instanciation:

    >>> cfg = TemplateDict({})

    :Setting values:

    Following exemples assume `cfg` is a :class:`incipyt._internal.templates.StringTemplate` around an empty `dict`.

    Bare values will be wrapped into :class:`incipyt._internal.templates.StringTemplate`
    automatically :

    >>> cfg["key"] = "{VARIABLE_NAME}"
    >>> print(cfg)
        TemplateDict(data={'key': StringTemplate(format_string={VARIABLE_NAME})})

    :class:`incipyt._internal.templates.Formattable` will be kept as-is:

    >>> cfg["key"] = a_formattable
    >>> print(cfg)
        TemplateDict(data={'key': a_formattable})

    Collections are supported as well:

    >>> cfg["key"] = ["{VARIABLE_NAME}"]
    >>> print(cfg)
        TemplateDict(data={'key': [StringTemplate("{VARIABLE_NAME}")]})

    >>> cfg["key"] = {"keyB": "{VARIABLE_NAME}"}
    >>> print(cfg)
        TemplateDict(data={'key': {'keyB': StringTemplate(format_string={VARIABLE_NAME})}})

    :Nested keys:

    To ease nested structure definition, the following syntax is supported:

    >>> cfg["keyA", "keyB"] = "{VARIABLE_NAME}"
    >>> print(cfg)
        TemplateDict(data={'keyA': {'keyB': StringTemplate(format_string={VARIABLE_NAME})}})

    :Multiple values:

    Instances of :class:`incipyt._internal.templates.ChoiceTemplate` will be
    created in case of value overrides. For instance, if `previous_value` is a
    :class:`incipyt._internal.templates.Formattable`:

    >>> cfg = TemplateDict({"key": previous_value})
    >>> cfg["key"] = "{VARIABLE_NAME}"
    >>> print(cfg)
        TemplateDict(data={'key': ChoiceTemplate({StringTemplate(format_string={VARIABLE_NAME}), previous_value})})

    If `previous_list` is a mutable sequence, any value not already present in
    it will be appended:

    >>> cfg = TemplateDict({"key": previous_list})
    >>> cfg["key"] = ["{VARIABLE_NAME}"]
    >>> cfg == {"key": previous_list + [StringTemplate("{VARIABLE_NAME}")]}
        True

    :Inplace union:

    The inplace union operator `|=` can be used to set multiple keys at once:

    >>> cfg = TemplateDict({})
    >>> cgf |= {"keyA": "{VARIABLE_NAME}", "keyB": "{OTHER_NAME}"}
    >>> print(cfg)
        TemplateDict(data={'keyA': StringTemplate(format_string={VARIABLE_NAME}), 'keyB': StringTemplate(format_string={OTHER_NAME})})
    """

    def __init__(self, data):
        """Proxy class around a provided mapping.

        This is itended to ease configuration templating.

        :param mapping: Existing mapping holding entries to wrap.
        :type mapping: :class:`collections.abc.MutableMapping`
        """
        self.data = data

    def __getitem__(self, keys):
        if not utils.is_nonstring_sequence(keys):
            keys = (keys,)

        config = self.data

        for key in keys:
            if key not in config:
                raise KeyError(f"Index [{', '.join(keys)}] does not exist.")
            config = config[key]

        if isinstance(config, abc.MutableMapping):
            return TemplateDict(config)
        elif utils.is_nonstring_sequence(config):
            return TemplateList(config)
        else:
            return config

    def __iter__(self):
        return iter(self.data)

    def __len__(self):
        return len(self.data)

    def __repr__(self):
        return utils.make_repr(self, "data")

    def __delitem__(self, key):
        raise NotImplementedError(
            f"{type(self)} do not support __delitem__, add-only dict-like."
        )

    def __setitem__(self, keys, value):
        if utils.is_nonstring_sequence(keys):
            config = self.data

            for key in keys[:-1]:
                if key not in config:
                    config[key] = {}
                config = config[key]

            self[keys[:-1]][keys[-1]] = value
            return

        if isinstance(value, abc.Mapping):
            if keys not in self.data:
                self.data[keys] = {}

            if utils.is_nonstring_sequence(self.data[keys]):
                raise TypeError(
                    f"{self.data[keys]}) is already a sequence, cannot set to a dict."
                )
            for k, v in value.items():
                TemplateDict(self.data[keys])[k] = v

        elif utils.is_nonstring_sequence(value):
            if keys not in self.data:
                self.data[keys] = []

            if isinstance(self.data[keys], abc.Mapping):
                raise TypeError(
                    f"{self.data[keys]} is already a mapping, cannot set to a list."
                )
            TemplateList(self.data[keys]).extend(value)

        else:
            if keys in self.data:
                self.data[keys] = ChoiceTemplate(value, self.data[keys])
            else:
                self.data[keys] = StringTemplate.wrap(value)

    def __ior__(self, other):
        self.update(other)
        return self


class TemplateList(abc.MutableSequence):
    """Proxy class around a provided mapping.

    This is itended to ease configuration templating.

    See :class:`incipyt._internal.templates.StringTemplate` for usage details.
    """

    def __init__(self, data):
        """Proxy class around a provided sequence.

        This is itended to ease configuration templating.

        :param mapping: Existing mapping holding entries to wrap.
        :type mapping: :class:`collections.abc.MutableSequence`
        """
        self.data = data

    def __getitem__(self, index):
        if utils.is_nonstring_sequence(self.data[index]):
            return TemplateList(self.data[index])
        elif isinstance(self.data[index], abc.MutableMapping):
            return TemplateDict(self.data[index])
        else:
            return self.data[index]

    def __len__(self):
        return len(self.data)

    def __repr__(self):
        return utils.make_repr(self, "data")

    def __eq__(self, other):
        return utils.attrs_eq(self, other, "data")

    def __setitem__(self, index, value):
        raise NotImplementedError(
            f"{type(self)} do not support __setitem__, add-only list-like."
        )

    def __delitem__(self, value):
        raise NotImplementedError(
            f"{type(self)} do not support del, add-only list-like."
        )

    def insert(self, index, value):
        if utils.is_nonstring_sequence(value):
            self.data.insert(index, [])
            TemplateList(self.data[index]).extend(value)
        elif isinstance(value, abc.Mapping):
            self.data.insert(index, {})
            TemplateDict(self.data[index]).update(value)
        else:
            new_value = StringTemplate.wrap(value)
            if new_value not in self.data:
                self.data.insert(index, new_value)


class FormatterEnviron(abc.Mapping):
    """Class wrapping an environ and providing an interface to render templates.

    It can be used to render template strings.
    """

    def __init__(self, sanitizer=None, value_error=True):
        """Class wrapping an environ and providing an interface to render templates.

        :param sanitizer: An optionnal callable to sanitize the values given (key, value) pairs.
        :type sanitizer: :class:`function` or `None`, optionnal
        :param environ: Consider empty string value as an error.
        :type environ: :class:`bool`, optional
        """
        self.data = project.environ
        self._keys = set()
        self._sanitizer = sanitizer
        self._value_error = value_error

    def __contains__(self, key):
        if key not in self.data:
            self.data.__getitem__(key)

        return True

    def __getitem__(self, key):
        # Call to inner dict __getitem__ will create missing keys
        value = self.data[key]
        if self._value_error and not value:
            raise ValueError

        return self._sanitizer(key, value) if self._sanitizer else value

    def __iter__(self):
        return iter(self._keys)

    def __len__(self):
        return len(self._keys)

    def keys(self):
        return iter(self)

    def values(self):
        return (self[key] for key in self)

    def items(self):
        return zip(self.keys(), self.values())

    def format(self, format_string, **kwargs):  # noqa: A003
        r"""Render a format string.

        Variables will be request from the underlying environ, and undefined
        variables will be created. If `self._value_error` is `True` and an
        empty variable will cause the whole render result to be `None`.

        Additional variables can be specified using keyword arguments. They
        will be added to the environ, hence they will override environ values.

        :param template: A format string whose keyword argument will be substituted.
        :type template: :class:`str`
        :param \**kwargs: Additional variables that will override environ variables.
        :type \**kwargs: :class:`str`, optionnal
        :return: The rendered template or `None` if a context variable is empty.
        :rtype: :class:`str` or `None`
        """

        self._keys = {item[1] for item in Formatter().parse(format_string) if item[1]}
        for key in self:
            if key in kwargs:
                self.data[key] = kwargs[key]

        with contextlib.suppress(ValueError):
            return format_string.format(**self)



from pytest import fixture, mark, raises

from incipyt._internal.dumpers import CfgIni, TextFile, Toml
from incipyt import project


@fixture
def reset_environ():
    project.environ.clear()


@mark.parametrize("dumper", (CfgIni, Toml, TextFile))
def test_format_path(dumper, reset_environ, tmp_path):
    dmp = dumper("{first}/{second}.ext")
    project.environ["first"] = project.EnvValue("folder", confirmed=True)
    project.environ["second"] = project.EnvValue("file", confirmed=True)
    dmp.commit(tmp_path)
    assert dmp.format_path() == tmp_path / "folder" / "file.ext"


@mark.parametrize("dumper", (CfgIni, Toml, TextFile))
def test_mkdir(dumper, reset_environ, tmp_path):
    dmp = dumper("folder/file")
    dmp.commit(tmp_path)
    dmp.mkdir()
    assert (tmp_path / "folder").is_dir()


@mark.parametrize("dumper", (CfgIni, Toml, TextFile))
def test_path_exists(dumper, reset_environ, tmp_path):
    (tmp_path / "file").touch()
    dmp = dumper("file")
    with raises(FileExistsError):
        dmp.commit(tmp_path)


@mark.parametrize(
    "dumper, data, res",
    [
        (
            CfgIni,
            {
                "section": {
                    "first": "1",
                    "second": {"one": "1", "two": "2"},
                    "third": ["one", "two"],
                }
            },
            (
                "[section]\n"
                "first = 1\n"
                "second = \n\tone = 1\n\ttwo = 2\n"
                "third = \n\tone\n\ttwo\n"
                "\n"
            ),
        ),
        (
            Toml,
            {
                "section": {
                    "first": "1",
                    "second": {"one": "1", "two": "2"},
                    "third": ["one", "two"],
                    "fourth": {"one": {"two": "2"}},
                    "fifth": [
                        {"one": "1.1", "two": "1.2"},
                        {"one": "2.1", "two": "2.2"},
                    ],
                }
            },
            (
                "[section]\n"
                'first = "1"\n'
                'third = [ "one", "two",]\n'
                "[[section.fifth]]\n"
                'one = "1.1"\n'
                'two = "1.2"\n'
                "\n"
                "[[section.fifth]]\n"
                'one = "2.1"\n'
                'two = "2.2"\n'
                "\n"
                "[section.second]\n"
                'one = "1"\n'
                'two = "2"\n'
                "\n"
                "[section.fourth.one]\n"
                'two = "2"\n'
            ),
        ),
        (
            TextFile,
            ["first", "second", "third"],
            "first\nsecond\nthird\n",
        ),
    ],
)
def test_dumpfile(dumper, data, res, reset_environ, tmp_path):
    dmp = dumper("file")
    dmp.commit(tmp_path)
    dmp.dump_in(data)
    assert (tmp_path / "file").read_text() == res



import os

from incipyt import commands, signals, tools


class Venv(tools.Tool):
    """Scripts to add virtualenv to :class:`incipyt.project._Structure`."""

    def add_to_structure(self):
        """Add venv configuration to `project.structure`, do nothing."""
        signals.vcs_ignore.emit(pattern=".env")

    def pre(self, workon):
        """Run `python -m venv .env`.

        :param workon: Work-on folder.
        :type workon: :class:`pathlib.Path`
        """
        env_path = workon / ".env"
        commands.venv([os.fspath(env_path)])
        commands.setenv_python_cmd(
            env_path.resolve() / ("Scripts" if os.name == "nt" else "bin") / "python"
        )



import os
import sys

from pytest import fixture, mark, raises

from incipyt import project, commands
from incipyt._internal.dumpers import TextFile, Toml
from incipyt._internal.templates import StringTemplate

from tests.utils import mock_stdin


class TestEnviron:
    @fixture
    def empty_environ(self, fake_process):
        fake_process.register_subprocess(["cmd", "arg"], stdout=["lineA", "lineB"])
        project.environ.clear()

    @fixture
    def simple_environ(self, fake_process):
        fake_process.register_subprocess(["cmd", "arg"], stdout=["lineA", "lineB"])
        project.environ.clear()
        project.environ["ONE"] = project.EnvValue("1", confirmed=True)

    @fixture
    def empty_auto_env(self, fake_process):
        fake_process.register_subprocess(["cmd", "arg"], stdout=["lineA", "lineB"])
        project.environ.clear()

    @fixture
    def simple_auto_env(self, fake_process):
        fake_process.register_subprocess(["cmd", "arg"], stdout=["lineA", "lineB"])
        project.environ.clear()
        project.environ["ONE"] = project.EnvValue("1", confirmed=True)

    @mark.parametrize(
        "env, input_values",
        (
            ("empty_environ", ["1"]),
            ("simple_environ", [""]),
            ("empty_auto_env", ["1"]),
            ("simple_auto_env", []),
        ),
    )
    def test_pull(self, env, input_values, monkeypatch, request):
        request.getfixturevalue(env)
        mock_stdin(monkeypatch, input_values)
        assert project.environ["ONE"] == "1"

    @mark.parametrize(
        "env",
        (
            "simple_environ",
            "simple_auto_env",
        ),
    )
    def test_push(self, env, request):
        request.getfixturevalue(env)
        with raises(ValueError):
            project.environ["ONE"] = "11"

    @mark.parametrize(
        "env, input_values",
        (
            ("simple_environ", [""]),
            ("simple_auto_env", []),
        ),
    )
    def test_update(self, env, input_values, monkeypatch, request):
        request.getfixturevalue(env)
        project.environ["ONE"] = project.EnvValue("11", update=True)
        mock_stdin(monkeypatch, input_values)
        assert project.environ["ONE"] == "11"

    @mark.parametrize(
        "env",
        (
            "empty_environ",
            "empty_auto_env",
        ),
    )
    def test_confirmed(self, env, request):
        request.getfixturevalue(env)
        project.environ["ONE"] = project.EnvValue("1", confirmed=True)
        assert project.environ["ONE"] == "1"

    @mark.parametrize(
        "env",
        (
            "empty_environ",
            "simple_environ",
            "empty_auto_env",
            "simple_auto_env",
        ),
    )
    def test_python_cmd(self, env, request):
        request.getfixturevalue(env)
        assert project.environ["PYTHON_CMD"] == sys.executable

    @mark.parametrize(
        "env",
        (
            "empty_environ",
            "empty_auto_env",
        ),
    )
    def test_iter(self, env, request):
        request.getfixturevalue(env)
        project.environ["TWO"] = project.EnvValue("2", confirmed=True)
        result = {key: project.environ[key] for key in project.environ}
        assert result == {"PYTHON_CMD": sys.executable, "TWO": "2"}

    @mark.parametrize(
        "env",
        (
            "empty_environ",
            "simple_environ",
            "empty_auto_env",
            "simple_auto_env",
        ),
    )
    def test_run(self, env, request):
        request.getfixturevalue(env)
        result = commands.run(["cmd", "arg"])
        assert result.stdout.decode() == f"lineA{os.linesep}lineB{os.linesep}"


class TestStructure:
    @fixture
    def reset_environ(self, fake_process):
        fake_process.register_subprocess(["cmd", "arg"], stdout=["lineA", "lineB"])
        project.environ.clear()
        project.environ["FOLDER_A"] = project.EnvValue("folderA", confirmed=True)
        project.environ["FOLDER_B"] = project.EnvValue("folderB", confirmed=True)
        project.environ["NAME_A"] = project.EnvValue("testA", confirmed=True)
        project.environ["NAME_B"] = project.EnvValue("testB", confirmed=True)
        project.environ["VALUE"] = project.EnvValue("1", confirmed=True)
        project.environ["CONTENT"] = project.EnvValue("text", confirmed=True)

    @fixture
    def reset_structure(self):
        project.structure.clear()
        project.structure.get_config_dict(Toml("{FOLDER_A}/{NAME_A}.toml"))[
            "section"
        ] = {"first": "{VALUE}"}
        project.structure.get_config_list(
            TextFile("{FOLDER_B}/{NAME_B}", sep="\n\n")
        ).append("{CONTENT}")

    def test_get_new_configuration(self, reset_structure):
        configuration = project.structure.get_config_dict(Toml("testC.toml"))
        assert configuration == {}

    def test_get_old_configuration(self, reset_structure):
        configuration = project.structure.get_config_dict(
            Toml("{FOLDER_A}/{NAME_A}.toml")
        )
        assert configuration == {"section": {"first": StringTemplate("{VALUE}")}}

    def test_mkdir(self, reset_structure, reset_environ, tmp_path):
        project.structure.mkdir(tmp_path)
        assert (tmp_path / "folderA").is_dir()
        assert (tmp_path / "folderB").is_dir()

    def test_commit(self, reset_structure, reset_environ, tmp_path):
        project.structure.mkdir(tmp_path)
        project.structure.commit()
        assert (
            tmp_path / "folderA" / "testA.toml"
        ).read_text() == '[section]\nfirst = "1"\n'
        assert (tmp_path / "folderB" / "testB").read_text() == "text\n\n"



import logging
import os
import pathlib
import sys
import warnings

import click

from incipyt import tools, project


logger = logging.getLogger(__name__)
DEFAULT_LOGGING_LEVEL = logging.WARNING
DEFAULT_FORMAT = "[%(levelname)s] %(message)s"
DEBUG_FORMAT = "%(asctime)s [%(levelname)s] <%(funcName)s> %(message)s"


@click.command(help="incipyt is a command-line tool that bootstraps a Python project.")
@click.argument(
    "folder",
    required=True,
    default=pathlib.Path(),
    type=click.Path(file_okay=False),
    callback=lambda _ctx, _param, _path: pathlib.Path(_path),
)
@click.option("-v", "--verbose", count=True)
@click.option("-s", "--silent", count=True)
@click.version_option()
@click.option(
    "--check-build",
    is_flag=True,
    help="Build the package after initialization of all files and folders.",
)
def main(folder, verbose, silent, check_build):
    log_level = DEFAULT_LOGGING_LEVEL - verbose * 10 + silent * 10
    setup_logging(max(logging.NOTSET, min(log_level, logging.CRITICAL)))

    if folder == pathlib.Path():
        if any(folder.resolve().iterdir()):
            raise click.BadArgumentUsage(f"FOLDER {folder.resolve()} is not empty.")
        project.environ["PROJECT_NAME"] = folder.resolve().name
    else:
        if (folder.is_absolute() and folder.is_dir() and any(folder.iterdir())) or (
            ("." / folder).is_dir() and any(("." / folder).resolve().iterdir())
        ):
            raise click.BadArgumentUsage(f"FOLDER {folder} is not empty.")
        project.environ["PROJECT_NAME"] = folder.name

    tools_to_install = [tools.Git(), tools.Venv(), tools.Setuptools(check_build)]

    for tool in tools_to_install:
        logger.info("Using %s", tool)
        tool.add_to_structure()

    logger.info("The project will be created at %s", folder.resolve())
    project.structure.mkdir(folder)

    for tool in tools_to_install:
        logger.info("Running pre-script for %s...", tool)
        tool.pre(folder)

    logger.info("Commit project structure.")
    project.structure.commit()

    for tool in tools_to_install:
        logger.info("Running post-script for %s...", tool)
        tool.post(folder)

    logger.info("All done.")


class ColoredFormatter(logging.Formatter):
    """Classic formatter with colored [LEVEL]."""

    colors = {10: (34, 49), 20: (32, 49), 30: (33, 49), 40: (31, 49), 50: (37, 41)}

    def format(self, record):  # noqa: A003, D102
        fg, bg = type(self).colors.get(record.levelno, (32, 49))
        record.levelname = f"\033[1;{fg}m\033[1;{bg}m{record.levelname}\033[0m"
        return super().format(record)


def supports_color(stream):
    """Determine if the given stream support colors."""
    platform_is_supported = True
    stream_is_a_tty = hasattr(stream, "isatty") and stream.isatty()
    return (platform_is_supported or "ANSICON" in os.environ) and stream_is_a_tty


def setup_logging(verbosity):
    """Set up the root logger.

    Replace the current root logger handlers by a new
    :class:`logging.StreamHandler` with the correct level and formatter.

    When the requested :param verbosity: is more verbose than
    :data:`logging.DEBUG` then the python warnings are logged too.

    :param verbosity int: the verbosity level to use on the root logger
    """
    root_logger = logging.getLogger("" if __name__ == "__main__" else __package__)
    stdout = logging.StreamHandler()
    f_cls = ColoredFormatter if supports_color(stdout.stream) else logging.Formatter
    f_format = DEFAULT_FORMAT if verbosity > logging.DEBUG else DEBUG_FORMAT
    stdout.formatter = f_cls(f_format)
    root_logger.handlers.clear()
    root_logger.handlers.append(stdout)
    root_logger.level = max(verbosity, logging.DEBUG)
    logger.level = root_logger.level - 10
    if verbosity < logging.DEBUG:
        logging.captureWarnings(True)
        warnings.filterwarnings("default")


# Remove '' and current working directory from the first entry of sys.path, if
# present to avoid using current directory in incipyt commands, when invoked as
# python -m incipyt <command>
if sys.path[0] in ("", os.getcwd()):
    sys.path.pop(0)

if __name__ == "__main__":
    sys.exit(main())
